@@ -1,15 +1,18 @@
 import logging
+import sys
 import time
 from enum import Enum
 
 import kubernetes
 import six
+from six import raise_from
 
 from dagster import DagsterInstance, check
+from dagster.core.errors import _add_inner_exception_for_py2
 from dagster.core.storage.pipeline_run import PipelineRunStatus
 
 DEFAULT_WAIT_TIMEOUT = 86400.0  # 1 day
-DEFAULT_WAIT_BETWEEN_ATTEMPTS = 1.0  # 1 second
+DEFAULT_WAIT_BETWEEN_ATTEMPTS = 10.0  # 10 seconds
 DEFAULT_JOB_POD_COUNT = 1  # expect job:pod to be 1:1 by default
 
 
@@ -22,10 +25,102 @@ class DagsterK8sError(Exception):
     pass
 
 
+class DagsterK8sTimeoutError(DagsterK8sError):
+    pass
+
+
+class DagsterK8sAPIRetryLimitExceeded(Exception):
+    def __init__(self, *args, **kwargs):
+        k8s_api_exception = check.inst_param(
+            kwargs.pop("k8s_api_exception"), "k8s_api_exception", Exception
+        )
+        original_exc_info = check.tuple_param(kwargs.pop("original_exc_info"), "original_exc_info")
+        max_retries = check.int_param(kwargs.pop("max_retries"), "max_retries")
+
+        check.invariant(original_exc_info[0] is not None)
+        msg = _add_inner_exception_for_py2(args[0], original_exc_info)
+        super(DagsterK8sAPIRetryLimitExceeded, self).__init__(
+            "Retry limit of {max_retries} exceeded: ".format(max_retries=max_retries) + msg,
+            *args[1:],
+            **kwargs
+        )
+
+        self.k8s_api_exception = check.opt_inst_param(
+            k8s_api_exception, "k8s_api_exception", Exception
+        )
+        self.original_exc_info = original_exc_info
+
+
+class DagsterK8sUnrecoverableAPIError(Exception):
+    def __init__(self, *args, **kwargs):
+        k8s_api_exception = check.inst_param(
+            kwargs.pop("k8s_api_exception"), "k8s_api_exception", Exception
+        )
+        original_exc_info = check.tuple_param(kwargs.pop("original_exc_info"), "original_exc_info")
+
+        check.invariant(original_exc_info[0] is not None)
+        msg = _add_inner_exception_for_py2(args[0], original_exc_info)
+        super(DagsterK8sUnrecoverableAPIError, self).__init__(msg, *args[1:], **kwargs)
+
+        self.k8s_api_exception = check.opt_inst_param(
+            k8s_api_exception, "k8s_api_exception", Exception
+        )
+        self.original_exc_info = original_exc_info
+
+
 class DagsterK8sPipelineStatusException(Exception):
     pass
 
 
+WHITELISTED_TRANSIENT_K8S_STATUS_CODES = [
+    429,  # Too many requests
+    504,  # Gateway timeout
+]
+
+
+def k8s_api_retry(
+    fn,
+    msg_fn=lambda: "Unexpected error encountered in Kubernetes API Client.",
+    max_retries=1,
+    timeout=10,
+):
+    check.callable_param(fn, "fn")
+
+    retries = max_retries
+    while retries > 0:
+        retries -= 1
+
+        try:
+            return fn()
+        except kubernetes.client.rest.ApiException as e:
+            # Only catch whitelisted ApiExceptions
+            status = e.status
+
+            # Check if the status code is generally whitelisted
+            whitelisted = status in WHITELISTED_TRANSIENT_K8S_STATUS_CODES
+
+            # If there are remaining retries, swallow the error
+            if whitelisted and retries > 0:
+                time.sleep(timeout)
+            elif whitelisted and retries == 0:
+                raise_from(
+                    DagsterK8sAPIRetryLimitExceeded(
+                        msg_fn(),
+                        k8s_api_exception=e,
+                        max_retries=max_retries,
+                        original_exc_info=sys.exc_info(),
+                    ),
+                    e,
+                )
+            else:
+                raise_from(
+                    DagsterK8sUnrecoverableAPIError(
+                        msg_fn(), k8s_api_exception=e, original_exc_info=sys.exc_info(),
+                    ),
+                    e,
+                )
+
+
 class KubernetesWaitingReasons:
     PodInitializing = "PodInitializing"
     ContainerCreating = "ContainerCreating"
@@ -66,10 +161,12 @@ class DagsterKubernetesClient:
         check.str_param(job_name, "job_name")
         check.str_param(namespace, "namespace")
 
-        pod_names = self.get_pod_names_for_job(job_name, namespace)
-
         try:
+            pod_names = self.get_pod_names_for_job(job_name, namespace)
+
             # Collect all the errors so that we can post-process before raising
+            pod_names = self.get_pod_names_for_job(job_name, namespace)
+
             errors = []
             try:
                 self.batch_api.delete_namespaced_job(name=job_name, namespace=namespace)
@@ -152,51 +249,62 @@ class DagsterKubernetesClient:
         job = None
         start = self.timer()
 
-        # Ensure we found the job that we launched
+        # Wait for job to launch
         while not job:
             if self.timer() - start > wait_timeout:
+                raise DagsterK8sTimeoutError(
+                    "Timed out while waiting for job {job_name}"
+                    " to launch".format(job_name=job_name)
+                )
 
-                raise DagsterK8sError("Timed out while waiting for job to launch")
+            # Get all jobs in the namespace and find the matching job
+            def _get_jobs_for_namespace():
+                jobs = self.batch_api.list_namespaced_job(namespace=namespace)
+                return next((j for j in jobs.items if j.metadata.name == job_name), None)
 
-            jobs = self.batch_api.list_namespaced_job(namespace=namespace)
-            job = next((j for j in jobs.items if j.metadata.name == job_name), None)
+            job = k8s_api_retry(_get_jobs_for_namespace, max_retries=3)
 
             if not job:
                 self.logger('Job "{job_name}" not yet launched, waiting'.format(job_name=job_name))
                 self.sleeper(wait_time_between_attempts)
 
-        # Wait for job completed status
+        # Wait for the job status to be completed. We check the status every
+        # wait_time_between_attempts seconds
         while True:
             if self.timer() - start > wait_timeout:
-                raise DagsterK8sError("Timed out while waiting for job to complete")
-
-            # See: https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.11/#jobstatus-v1-batch
-            status = self.batch_api.read_namespaced_job_status(job_name, namespace=namespace).status
-
-            if status.failed and status.failed > 0:
-                pods = self.core_api.list_namespaced_pod(
-                    label_selector="job-name=={}".format(job_name), namespace=namespace
-                )
-                logs = {}
-                for pod in pods.items:
-                    pod_name = pod.metadata.name
-                    try:
-                        logs[pod_name] = self.core_api.read_namespaced_pod_log(
-                            name=pod_name, namespace=namespace
-                        )
-                    except kubernetes.client.rest.ApiException as e:
-                        logs[pod_name] = e
-
-                raise DagsterK8sError(
-                    "Encountered failed job pods with status: {}, and logs: {}".format(status, logs)
+                raise DagsterK8sTimeoutError(
+                    "Timed out while waiting for job {job_name}"
+                    " to complete".format(job_name=job_name)
                 )
 
-            # done waiting for pod completion
+            # Reads the status of the specified job. Returns a V1Job object that
+            # we need to read the status off of.
+            status = None
+
+            def _get_job_status():
+                job = self.batch_api.read_namespaced_job_status(job_name, namespace=namespace)
+                return job.status
+
+            status = k8s_api_retry(_get_job_status, max_retries=3)
+
+            # status.succeeded represents the number of pods which reached phase Succeeded.
             if status.succeeded == num_pods_to_wait_for:
                 break
 
+            # status.failed represents the number of pods which reached phase Failed.
+            if status.failed and status.failed > 0:
+                raise DagsterK8sError(
+                    "Encountered failed job pods for job {job_name} with status: {status}".format(
+                        job_name=job_name, status=status
+                    )
+                )
+
             if instance and run_id:
-                pipeline_run_status = instance.get_run_by_id(run_id).status
+                pipeline_run = instance.get_run_by_id(run_id)
+                if not pipeline_run:
+                    raise DagsterK8sPipelineStatusException()
+
+                pipeline_run_status = pipeline_run.status
                 if pipeline_run_status != PipelineRunStatus.STARTED:
                     raise DagsterK8sPipelineStatusException()
 
